package org.ow2.kerneos.flex;

import org.apache.felix.ipojo.ComponentInstance;
import org.apache.felix.ipojo.Factory;
import org.apache.felix.ipojo.annotations.Bind;
import org.apache.felix.ipojo.annotations.Component;
import org.apache.felix.ipojo.annotations.Instantiate;
import org.apache.felix.ipojo.annotations.Invalidate;
import org.apache.felix.ipojo.annotations.Provides;
import org.apache.felix.ipojo.annotations.Requires;
import org.apache.felix.ipojo.annotations.Unbind;
import org.apache.felix.ipojo.annotations.Validate;
import org.granite.gravity.osgi.adapters.ea.EAConstants;
import org.osgi.service.cm.Configuration;
import org.osgi.service.cm.ConfigurationAdmin;
import org.ow2.kerneos.common.KerneosConstants;
import org.ow2.kerneos.common.config.generated.Service;
import org.ow2.kerneos.common.config.generated.SwfModule;
import org.ow2.kerneos.common.service.KerneosApplication;
import org.ow2.kerneos.common.service.KerneosModule;
import org.ow2.kerneos.common.service.KerneosModuleFragment;
import org.ow2.kerneos.login.KerneosSession;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;
import java.util.Collection;
import java.util.Collections;
import java.util.Dictionary;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Map;

/**
 * The core of Flex Kerneos.
 */
@Component
@Instantiate
@Provides
public class Core implements ICore {
    private static final String KERNEOS_SESSION_KEY = "KERNEOS-SESSION-";

    /**
     * The logger.
     */
    private static final Log LOGGER = LogFactory.getLog(Core.class);

    private Configuration granite, gravity;

    private Map<String, KerneosApplication> applications = new HashMap<String, KerneosApplication>();
    private Map<String, ComponentInstance> applicationHttpServices = new HashMap<String, ComponentInstance>();
    private Map<String, KerneosModule> modules = new HashMap<String, KerneosModule>();
    private Map<String, Map<String, KerneosModuleFragment>> modulesFragments = new HashMap<String, Map<String, KerneosModuleFragment>>();

    @Requires
    private ConfigurationAdmin configurationAdmin;

    @Requires(filter = "(factory.name=org.ow2.kerneos.flex.FlexHttpService)")
    private Factory flexHttpServiceFactory;

    /**
     * Constructor.
     * Avoid direct component instantiation.
     */
    private Core() {

    }

    /**
     * Called when all the component dependencies are met.
     *
     * @throws IOException an issue occurs during the validation
     */
    @Validate
    private synchronized void start() throws Exception {
        LOGGER.debug("Start FlexCore");

        // Gravity Configurations
        {
            Dictionary properties = new Hashtable();
            properties.put("id", FlexConstants.GRAVITY_SERVICE);
            properties.put("messageTypes", "flex.messaging.messages.AsyncMessage");
            properties.put("defaultAdapter", EAConstants.ADAPTER_ID);
            gravity = configurationAdmin.createFactoryConfiguration(org.granite.config.flex.Service.class.getName(),
                    null);
            gravity.update(properties);
        }

        // Granite Configurations
        {
            Dictionary properties = new Hashtable();
            properties.put("id", FlexConstants.GRANITE_SERVICE);
            granite = configurationAdmin.createFactoryConfiguration(org.granite.config.flex.Service.class.getName(),
                    null);
            granite.update(properties);
        }

        // Start instances
        for (ComponentInstance instance : applicationHttpServices.values()) {
            instance.start();
        }
    }

    /**
     * Called when all the component dependencies aren't met anymore.
     *
     * @throws IOException an issue occurs during the validation
     */
    @Invalidate
    private synchronized void stop() throws IOException {
        LOGGER.debug("Stop FlexCore");

        // Stop instances
        for (ComponentInstance instance : applicationHttpServices.values()) {
            instance.stop();
        }

        // Dispose configurations
        gravity.delete();
        granite.delete();
    }


    /**
     * Optional Kerneos Application binding.
     * Create a new FlexHttpService for each Kerneos Application.
     *
     * @param application the new Kerneos Application
     */
    @Bind(aggregate = true, optional = true)
    private synchronized void bindKerneosApplication(KerneosApplication application) {
        applications.put(application.getId(), application);
        try {
            LOGGER.debug("Create FlexHttpService for " + application.getId());
            Dictionary properties = new Hashtable();
            Dictionary filters = new Hashtable();
            filters.put(FlexHttpService.APPLICATION, "(" + KerneosApplication.ID + "=" + application.getId() + ")");
            properties.put("requires.filters", filters);
            ComponentInstance instance = flexHttpServiceFactory.createComponentInstance(properties);
            applicationHttpServices.put(application.getId(), instance);
            instance.start();
        } catch (Exception e) {
            LOGGER.error("Can't create FlexHttpService for " + application.getId());
        }
    }

    /**
     * Optional Kerneos Application binding.
     * Remove the FlexHttpService associated with the Kerneos Application.
     *
     * @param application the old Kerneos Application
     */
    @Unbind
    private void unbindKerneosApplication(KerneosApplication application) {
        removeKerneosApplication(application.getId());
    }


    private synchronized void removeKerneosApplication(String applicationId) {
        applications.remove(applicationId);
        ComponentInstance instance = applicationHttpServices.remove(applicationId);
        if (instance != null) {
            instance.dispose();
            LOGGER.debug("Destroy FlexHttpService for " + applicationId);
        }
    }


    /**
     * Optional Kerneos Module binding.
     *
     * @param module the new Kerneos Module
     */
    @Bind(aggregate = true, optional = true)
    private synchronized void bindKerneosModule(KerneosModule module) {
        modules.put(module.getId(), module);
        if (!modulesFragments.containsKey(module.getId())) {
            modulesFragments.put(module.getId(), new HashMap<String, KerneosModuleFragment>());
        }
    }

    /**
     * Optional Kerneos Module binding.
     *
     * @param module the old Kerneos Module
     */
    @Unbind
    private synchronized void unbindKerneosModule(KerneosModule module) {
        removeKerneosModule(module.getId());
    }

    private synchronized void removeKerneosModule(String applicationId) {
        KerneosModule module = modules.remove(applicationId);
        modulesFragments.get(module.getId()).clear();
        modulesFragments.remove(module.getId());
    }

    /**
     * Update the current flex context.
     *
     * @param request  the HTTP request associated with the context
     * @param response the HTTP response associated with the context
     */
    @Override
    public void updateContext(HttpServletRequest request, HttpServletResponse response) {
        FlexContext flexContext = FlexContext.getCurrent();

        flexContext.setRequest(request);
        flexContext.setResponse(response);

        String path = request.getRequestURI();

        // Remove part before application (compatible felix + war)
        int index = path.indexOf(flexContext.getApplication().getConfiguration().getApplicationUrl());
        path = path.substring(index + flexContext.getApplication().getConfiguration().getApplicationUrl().length());

        // Get the module and the associated path
        KerneosModule currentModule = null;
        KerneosModuleFragment currentModuleFragment = null;
        if (path != null && path.startsWith(KerneosConstants.KERNEOS_MODULE_URL)) {
            path = path.substring(KerneosConstants.KERNEOS_MODULE_URL.length());
            int sep = path.indexOf("/");
            if (sep != -1) {
                synchronized (this) {
                    currentModule = modules.get(path.substring(0, sep));
                }
                path = path.substring(sep);
            }
            if (currentModule != null && path.startsWith(KerneosConstants.KERNEOS_MODULE_FRAGMENT_URL)) {
                path = path.substring(KerneosConstants.KERNEOS_MODULE_FRAGMENT_URL.length());
                sep = path.indexOf("/");
                if (sep != -1) {
                    synchronized (this) {
                        currentModuleFragment = modulesFragments.get(currentModule.getId()).get(path.substring(0, sep));
                    }
                    path = path.substring(sep);
                }
            }

        }

        flexContext.setModuleFragment(currentModuleFragment);
        flexContext.setModule(currentModule);
        flexContext.setPath(path);
        flexContext.setService(null);
        flexContext.setMethod(null);

        // Get or create a session
        KerneosSession session = null;
        String sessionKey = KERNEOS_SESSION_KEY + flexContext.getApplication().getId();
        Object obj = request.getSession().getAttribute(sessionKey);
        if (obj == null || !(obj instanceof KerneosSession)) {
            session = flexContext.getApplication().getLoginManager().newSession();
            if (session.getRoles() != null) {
                session.setRoles(flexContext.getApplication().getRolesManager().resolve(session.getRoles()));
            }
            request.getSession().setAttribute(sessionKey, session);
        } else {
            session = (KerneosSession) obj;
        }
        KerneosSession.setCurrent(session);
    }

    /**
     * Update the current flex context.
     *
     * @param destination the new destination associated with the context
     * @param method      the new method associated with the context
     */
    @Override
    public void updateContext(String destination, String method) {
        FlexContext flexContext = FlexContext.getCurrent();
        flexContext.setPath(null);
        flexContext.setMethod(method);
        KerneosModule module = null;
        KerneosModuleFragment moduleFragment = null;
        String service = null;
        synchronized (this) {
            for (KerneosModule amodule : modules.values()) {
                if (amodule.getConfiguration() instanceof SwfModule) {
                    SwfModule configuration = (SwfModule) amodule.getConfiguration();
                    for (Service aservice : configuration.getServices()) {
                        if (aservice.getDestination().equals(destination)) {
                            service = aservice.getId();
                            module = amodule;
                            break;
                        }
                    }
                }
                if (module != null) {
                    break;
                }
                for (KerneosModuleFragment fragment : modulesFragments.get(amodule.getId()).values()) {
                    for (Service fragmentService : fragment.getConfiguration().getServices()) {
                        if (fragmentService.getDestination().equals(destination)) {
                            service = fragmentService.getId();
                            moduleFragment = fragment;
                            module = amodule;
                            break;
                        }
                    }
                }
                if (moduleFragment != null) {
                    break;
                }
            }
        }
        if (module != null) {
            flexContext.setModule(module);
            flexContext.setService(service);
            flexContext.setModuleFragment(moduleFragment);
        }
    }

    /**
     * Optional Kerneos Module Fragment binding.
     *
     * @param moduleFragment the new Kerneos Module Fragment
     */
    @Bind(aggregate = true, optional = true)
    private synchronized void bindKerneosModuleFragment(KerneosModuleFragment moduleFragment) {
        if (!modulesFragments.containsKey(moduleFragment.getModuleId())) {
            modulesFragments.put(moduleFragment.getModuleId(), new HashMap<String, KerneosModuleFragment>());
        }
        modulesFragments.get(moduleFragment.getModuleId()).put(moduleFragment.getId(), moduleFragment);
    }

    /**
     * Optional Kerneos Module Fragment binding.
     *
     * @param moduleFragment the old Kerneos Module Fragment
     */
    @Unbind
    private synchronized void unbindKerneosModuleFragment(KerneosModuleFragment moduleFragment) {
        if (modulesFragments.containsKey(moduleFragment.getModuleId())) {
            modulesFragments.get(moduleFragment.getModuleId()).remove(moduleFragment.getId());
        }
    }

    @Override
    public KerneosModuleFragment getModuleFragment(String moduleId, String fragmentId) {
        if (!modulesFragments.containsKey(moduleId)) {
            LOGGER.warn("Could not retrieve Module Fragment {0} from Module {1} because Module {1} doesn't exist",
                    fragmentId, moduleId);
            return null;
        }
        return modulesFragments.get(moduleId).get(fragmentId);
    }

    @Override
    public Collection<KerneosModuleFragment> getModuleFragments(String moduleId) {
        if (!modulesFragments.containsKey(moduleId)) {
            LOGGER.warn("Could not retrieve Module Fragments from Module {0} because Module {0} doesn't exist",
                    moduleId);
            return Collections.emptySet();
        }
        return modulesFragments.get(moduleId).values();
    }

    @Override
    public KerneosModule getModule(String moduleId) {
        return modules.get(moduleId);
    }

    @Override
    public KerneosApplication getApplication(String applicationId) {
        return applications.get(applicationId);
    }

    @Override
    public Collection<KerneosModule> getModules() {
        return modules.values();
    }

    @Override
    public Collection<KerneosApplication> getApplications() {
        return applications.values();
    }

}
